#include <iostream>
#include <chrono>
#include <fstream>
#include <memory>
#include <numeric>
#include <cassert>
#include <vector>
#include <unistd.h>
#include <NvInfer.h>
#include <cuda_runtime_api.h>

struct TrtDeleter
{
    template <typename T>
    void operator()(T *obj) const
    {
        if (obj)
        {
            obj->destroy();
        }
    }
};

template <typename T>
using TrtUniquePtr = std::unique_ptr<T, TrtDeleter>;

class Logger : public nvinfer1::ILogger
{
public:
    Logger(nvinfer1::ILogger::Severity severity = nvinfer1::ILogger::Severity::kWARNING)
        : m_severity(severity)
    {
    }

    void log(nvinfer1::ILogger::Severity severity, const char *msg)
    {
        if (severity <= m_severity)
        {
            std::cout << "[TRT] " << std::string(msg) << std::endl;
        }
    }

private:
    nvinfer1::ILogger::Severity m_severity;
};

Logger logger;

std::shared_ptr<nvinfer1::ICudaEngine> create_engine(const std::string &model_path, Logger &logger)
{
    std::ifstream model_file(model_path, std::ios::in | std::ios::binary);
    assert(!model_file.fail());

    model_file.seekg(0, std::ios::end);
    const int model_size = model_file.tellg();
    model_file.clear();
    model_file.seekg(0, std::ios::beg);
    std::vector<char> model_buffer(model_size);
    model_file.read(&model_buffer[0], model_size);
    model_file.close();

    auto runtime = TrtUniquePtr<nvinfer1::IRuntime>(nvinfer1::createInferRuntime(logger));
    auto engine = std::shared_ptr<nvinfer1::ICudaEngine>(runtime->deserializeCudaEngine(
                                                             (void *)&model_buffer[0], model_size, nullptr),
                                                         TrtDeleter());
    return engine;
}

int main(int argc, char **argv)
{
    std::cout << "[TRT] loading model..." << std::endl;
    auto engine = create_engine("../yolov5s.trt", logger);
    auto context = TrtUniquePtr<nvinfer1::IExecutionContext>(engine->createExecutionContext());

    nvinfer1::Dims i_dims = static_cast<nvinfer1::Dims &&>(engine->getBindingDimensions(engine->getBindingIndex("images")));
    nvinfer1::Dims o_dims = static_cast<nvinfer1::Dims &&>(engine->getBindingDimensions(engine->getBindingIndex("output")));
    const int i_size = std::accumulate(i_dims.d, i_dims.d + i_dims.nbDims, 1, std::multiplies<int64_t>());
    const int o_size = std::accumulate(o_dims.d, o_dims.d + o_dims.nbDims, 1, std::multiplies<int64_t>());

    std::vector<void *> buffers(2);
    assert(cudaError::cudaSuccess == cudaMalloc((void **)&buffers[0], i_size * sizeof(float)));
    assert(cudaError::cudaSuccess == cudaMalloc((void **)&buffers[1], o_size * sizeof(float)));

    std::cout << "[TRT] start continuous executeV2 for 1000 times..." << std::endl;
    double duration = 0.0;
    for (int i = 0; i < 1000; ++i)
    {
        auto start = std::chrono::steady_clock::now();
        context->executeV2(&buffers[0]);
        auto end = std::chrono::steady_clock::now();
        duration += std::chrono::duration<double, std::milli>(end - start).count();
    }
    duration /= 1000.;
    std::cout << "[TRT] continuous executeV2 mean time: " << duration << " ms" << std::endl;

    std::cout << "[TRT] start non-continuous executeV2 for 1000 times..." << std::endl;
    duration = 0.0;
    for (int i = 0; i < 1000; ++i)
    {
        auto start = std::chrono::steady_clock::now();
        context->executeV2(&buffers[0]);
        auto end = std::chrono::steady_clock::now();
        duration += std::chrono::duration<double, std::milli>(end - start).count();
        usleep(100 * 1000);
    }
    duration /= 1000.;
    std::cout << "[TRT] non-continuous executeV2 mean time: " << duration << " ms" << std::endl;

    cudaFree(buffers[0]);
    cudaFree(buffers[1]);

    return 0;
}
